@@ -53,9 +53,6 @@
 #include "chxj_chtml20.h"
 #include "chxj_chtml30.h"
 #include "chxj_jhtml.h"
-#ifdef USE_MYSQL_COOKIE
-#include "chxj_mysql.h"
-#endif
 
 #include "chxj_img_conv_format.h"
 #include "chxj_qr_code.h"
@@ -64,6 +61,9 @@
 #include "chxj_cookie.h"
 #include "chxj_url_encode.h"
 #include "chxj_str_util.h"
+#if defined(USE_MYSQL_COOKIE)
+#  include "chxj_mysql.h"
+#endif
 
 
 #define CHXJ_VERSION_PREFIX PACKAGE_NAME "/"
@@ -199,6 +199,7 @@ chxj_exchange(request_rec *r, const char** src, apr_size_t* len, device_table *s
   mod_chxj_config     *dconf; 
   chxjconvrule_entry  *entryp;
 
+  DBG(r,"start of chxj_exchange() input:[%.*s]", (int)*len, *src);
   dst  = apr_pstrcat(r->pool, (char*)*src, NULL);
 
   dconf = ap_get_module_config(r->per_dir_config, &chxj_module);
@@ -219,13 +220,13 @@ chxj_exchange(request_rec *r, const char** src, apr_size_t* len, device_table *s
     user_agent = (char*)apr_table_get(r->headers_in, HTTP_USER_AGENT);
 
   DBG(r,"User-Agent:[%s]", user_agent);
-  DBG(r, "start chxj_exchange()");
   DBG(r,"content type is %s", r->content_type);
 
 
-  if (*(char*)r->content_type == 't' 
-  && strncmp(r->content_type, "text/html",   9) != 0) {
-    DBG(r,"content type is %s", r->content_type);
+  if (! STRNCASEEQ('t','T', "text/html", r->content_type, sizeof("text/html"))
+  &&  ! STRNCASEEQ('a','A', "application/xhtml+xml", r->content_type, sizeof("application/xhtml+xml"))) {
+    DBG(r,"no convert. content type is %s", r->content_type);
+    DBG(r,"end of chxj_exchange()");
     return (char*)*src;
   }
 
@@ -287,7 +288,7 @@ chxj_exchange(request_rec *r, const char** src, apr_size_t* len, device_table *s
   }
   dst[*len] = 0;
 
-  DBG(r, "end chxj_exchange()");
+  DBG(r, "end of chxj_exchange()");
 
   return dst;
 }
@@ -618,14 +619,14 @@ chxj_output_filter(ap_filter_t *f, apr_bucket_brigade *bb)
   apr_bucket*         b;
   const char*         data;
   char*               contentLength;
-  char*               user_agent;
+  char*               user_agent = NULL;
   apr_size_t          len;
-  mod_chxj_ctx*       ctx;
+  mod_chxj_ctx*       ctx = (mod_chxj_ctx *)f->ctx;
   cookie_t*           cookie;
   char*               location_header;
   mod_chxj_config*    dconf;
-  chxjconvrule_entry  *entryp;
-  device_table        *spec;
+  chxjconvrule_entry  *entryp = NULL;
+  device_table        *spec = NULL;
 
 
 
@@ -642,6 +643,7 @@ chxj_output_filter(ap_filter_t *f, apr_bucket_brigade *bb)
   if (r->content_type) {
     if (! STRNCASEEQ('t','T',"text/html",r->content_type, sizeof("text/html")-1)
     &&  ! STRNCASEEQ('t','T',"text/xml", r->content_type, sizeof("text/xml")-1)
+    &&  ! STRNCASEEQ('a','A',"application/xhtml+xml", r->content_type, sizeof("application/xhtml+xml")-1)
     &&  ! (STRNCASEEQ('i','I',"image/",  r->content_type, sizeof("image/") -1)
           && ( STRCASEEQ('j','J',"jpeg",            &r->content_type[6])         /* JPEG */
             || STRCASEEQ('j','J',"jp2",             &r->content_type[6])         /* JPEG2000 */
@@ -664,22 +666,69 @@ chxj_output_filter(ap_filter_t *f, apr_bucket_brigade *bb)
   }
 
   dconf      = ap_get_module_config(r->per_dir_config, &chxj_module);
-  entryp     = chxj_apply_convrule(r, dconf->convrules);
   user_agent = (char*)apr_table_get(r->headers_in, HTTP_USER_AGENT);
-  spec       = chxj_specified_device(r, user_agent);
+  if (ctx && ctx->entryp) entryp = ctx->entryp;
+  else                    entryp = chxj_apply_convrule(r, dconf->convrules);
+  if (ctx && ctx->spec)   spec   = ctx->spec;
+  else                    spec   = chxj_specified_device(r, user_agent);
 
 
   for (b = APR_BRIGADE_FIRST(bb);
        b != APR_BRIGADE_SENTINEL(bb); 
        b = APR_BUCKET_NEXT(b)) {
 
+    if (apr_bucket_read(b, &data, &len, APR_BLOCK_READ) == APR_SUCCESS) {
+      DBG(r, "read data[%.*s]",(int)len, data);
+
+      if (f->ctx == NULL) {
+        /*--------------------------------------------------------------------*/
+        /* Start                                                              */
+        /*--------------------------------------------------------------------*/
+        DBG(r, "new context");
+        ctx = (mod_chxj_ctx*)apr_palloc(r->pool, sizeof(mod_chxj_ctx));
+        if (len > 0) {
+          ctx->buffer = apr_palloc(r->pool, len);
+          memcpy(ctx->buffer, data, len);
+        }
+        else {
+          ctx->buffer = apr_palloc(r->pool, 1);
+          ctx->buffer = '\0';
+        }
+        ctx->len = len;
+        f->ctx = (void*)ctx;
+        ctx->entryp = entryp;
+        ctx->spec   = spec;
+      }
+      else {
+        /*--------------------------------------------------------------------*/
+        /* append data                                                        */
+        /*--------------------------------------------------------------------*/
+        char* tmp;
+        DBG(r, "append data start");
+        ctx = (mod_chxj_ctx*)f->ctx;
+
+        if (len > 0) {
+          tmp = apr_palloc(r->pool, ctx->len);
+          memcpy(tmp, ctx->buffer, ctx->len);
+
+          ctx->buffer = apr_palloc(r->pool, ctx->len + len);
+
+          memcpy(ctx->buffer, tmp, ctx->len);
+          memcpy(&ctx->buffer[ctx->len], data, len);
+
+          ctx->len += len;
+        }
+        DBG(r, "append data end");
+      }
+    }
+
     if (APR_BUCKET_IS_EOS(b)) {
 
       DBG(r, "eos");
       /*----------------------------------------------------------------------*/
       /* End Of File                                                          */
       /*----------------------------------------------------------------------*/
-      if (f->ctx) {
+      if (ctx) {
 
         ctx = (mod_chxj_ctx*)f->ctx;
 
@@ -687,8 +736,8 @@ chxj_output_filter(ap_filter_t *f, apr_bucket_brigade *bb)
 
         if (spec->html_spec_type != CHXJ_SPEC_UNKNOWN 
             && r->content_type 
-            && *(char*)r->content_type == 't' 
-            && strncmp(r->content_type, "text/html",   9) == 0) {
+            && (STRNCASEEQ('a','A',"application/xhtml+xml", r->content_type, sizeof("application/xhtml+xml"))
+            ||  STRNCASEEQ('t','T',"text/html", r->content_type, sizeof("text/html")))) {
 
           if (ctx->len) {
             char* tmp;
@@ -869,49 +918,6 @@ chxj_output_filter(ap_filter_t *f, apr_bucket_brigade *bb)
         return rv;
       }
     }
-    else
-    if (apr_bucket_read(b, &data, &len, APR_BLOCK_READ) == APR_SUCCESS) {
-      DBG(r, "read data[%.*s]",(int)len, data);
-
-      if (f->ctx == NULL) {
-        /*--------------------------------------------------------------------*/
-        /* Start                                                              */
-        /*--------------------------------------------------------------------*/
-        DBG(r, "new context");
-        ctx = (mod_chxj_ctx*)apr_palloc(r->pool, sizeof(mod_chxj_ctx));
-        if (len > 0) {
-          ctx->buffer = apr_palloc(r->pool, len);
-          memcpy(ctx->buffer, data, len);
-        }
-        else {
-          ctx->buffer = apr_palloc(r->pool, 1);
-          ctx->buffer = '\0';
-        }
-        ctx->len = len;
-        f->ctx = (void*)ctx;
-      }
-      else {
-        /*--------------------------------------------------------------------*/
-        /* append data                                                        */
-        /*--------------------------------------------------------------------*/
-        char* tmp;
-        DBG(r, "append data start");
-        ctx = (mod_chxj_ctx*)f->ctx;
-
-        if (len > 0) {
-          tmp = apr_palloc(r->pool, ctx->len);
-          memcpy(tmp, ctx->buffer, ctx->len);
-
-          ctx->buffer = apr_palloc(r->pool, ctx->len + len);
-
-          memcpy(ctx->buffer, tmp, ctx->len);
-          memcpy(&ctx->buffer[ctx->len], data, len);
-
-          ctx->len += len;
-        }
-        DBG(r, "append data end");
-      }
-    }
   }
   apr_brigade_destroy(bb);
 
@@ -937,29 +943,31 @@ chxj_input_filter(ap_filter_t*        f,
                  apr_read_type_e      block,
                  apr_off_t            readbytes)
 {
-  request_rec*        r;
+  request_rec         *r;
   apr_status_t        rv;
-  conn_rec*           c;
-  apr_bucket*         b;
+  conn_rec            *c;
+  apr_bucket          *b;
   /*--------------------------------------------------------------------------*/
   /* It is the brigade area for output                                        */
   /*--------------------------------------------------------------------------*/
-  apr_bucket_brigade* ibb;            
+  apr_bucket_brigade  *ibb;            
   /*--------------------------------------------------------------------------*/
   /* It is the brigade area for input                                         */
   /*--------------------------------------------------------------------------*/
-  apr_bucket_brigade* obb;
+  apr_bucket_brigade  *obb;
   apr_size_t          len;
-  apr_bucket*         tmp_heap;
-  apr_bucket*         eos;
-  const char*         data;
-  char*               data_bucket;
-  char*               data_brigade;
-  char*               content_type;
-  device_table*       spec ;
-  char*               user_agent;
-  mod_chxj_config*    dconf;
-  chxjconvrule_entry* entryp;
+  apr_bucket          *tmp_heap;
+  apr_bucket          *eos;
+  const char          *data;
+  char                *data_bucket;
+  char                *data_brigade;
+  char                *content_type;
+  device_table        *spec = NULL;
+  char                *user_agent = NULL;
+  mod_chxj_config     *dconf;
+  chxjconvrule_entry  *entryp = NULL;
+  mod_chxj_ctx        *ctx = (mod_chxj_ctx *)f->ctx;
+  int                 eos_flag = 0;
 
   r = f->r;
   c = r->connection;
@@ -974,7 +982,7 @@ chxj_input_filter(ap_filter_t*        f,
 
   content_type = (char*)apr_table_get(r->headers_in, "Content-Type");
   if (content_type 
-  && strncasecmp("multipart/form-data", content_type, 19) == 0) {
+      && strncasecmp("multipart/form-data", content_type, 19) == 0) {
 
     DBG(r, "detect multipart/form-data");
     ap_remove_input_filter(f);
@@ -983,8 +991,14 @@ chxj_input_filter(ap_filter_t*        f,
   }
 
   dconf = ap_get_module_config(r->per_dir_config, &chxj_module);
+  user_agent = (char*)apr_table_get(r->headers_in, "User-Agent");
+
+  if (ctx && ctx->entryp) entryp = ctx->entryp;
+  else                    entryp = chxj_apply_convrule(r, dconf->convrules);
+
+  if (ctx && ctx->spec) spec = ctx->spec;
+  else                  spec = chxj_specified_device(r, user_agent);
 
-  entryp = chxj_apply_convrule(r, dconf->convrules);
   if (!entryp || !(entryp->action & CONVRULE_ENGINE_ON_BIT)) {
     DBG(r,"EngineOff");
 
@@ -992,8 +1006,6 @@ chxj_input_filter(ap_filter_t*        f,
     return ap_get_brigade(f->next, bb, mode, block, readbytes);
   }
 
-  user_agent = (char*)apr_table_get(r->headers_in, "User-Agent");
-  spec = chxj_specified_device(r, user_agent);
 
   switch(spec->html_spec_type) {
   case CHXJ_SPEC_Chtml_1_0:
@@ -1017,10 +1029,20 @@ chxj_input_filter(ap_filter_t*        f,
     DBG(r, "ap_get_brigade() failed");
     return rv;
   }
+  if (!ctx) {
+    ctx = apr_palloc(r->pool, sizeof(*ctx));
+    memset(ctx, 0, sizeof(*ctx));
+    if ((rv = apr_pool_create(&ctx->pool, r->pool)) != APR_SUCCESS) {
+      ERR(r, "failed: new pool create. rv:[%d]", rv);
+      return rv;
+    }
+    ctx->entryp = entryp;
+    ctx->spec = spec;
+    ctx->buffer = apr_palloc(ctx->pool, 1);
+    ctx->buffer[0] = 0;
+    f->ctx = ctx;
+  }
 
-#if 0
-  APR_BRIGADE_FOREACH(b, ibb) {
-#endif
   for (b =  APR_BRIGADE_FIRST(ibb); 
        b != APR_BRIGADE_SENTINEL(ibb);
        b =  APR_BUCKET_NEXT(b)) {
@@ -1032,50 +1054,60 @@ chxj_input_filter(ap_filter_t*        f,
     }
 
     if (data != NULL) {
-      data_bucket = apr_palloc(r->pool, len+1);
+      ctx->len += len;
+      data_bucket = apr_palloc(ctx->pool, len+1);
       memset((void*)data_bucket, 0, len+1);
       memcpy(data_bucket, data, len);
       DBG(r, "(in)POSTDATA:[%s]", data_bucket);
-  
-      data_brigade = apr_pstrcat(r->pool, data_brigade, data_bucket, NULL);
+      ctx->buffer = apr_pstrcat(ctx->pool, ctx->buffer, data_bucket, NULL);
     }
-
     if (APR_BUCKET_IS_EOS(b)) {
+      DBG(r, "eos");
+      eos_flag = 1;
       break;
     }
   }
   apr_brigade_cleanup(ibb);
 
-
-  len = strlen(data_brigade);
-  if (len == 0) {
+  if (ctx->len == 0) {
     DBG(r,"data_brigade length is 0");
     DBG(r,"end of chxj_input_filter()");
     ap_remove_input_filter(f);
     return ap_get_brigade(f->next, bb, mode, block, readbytes);
   }
 
-  data_brigade = chxj_input_convert(
-    r, 
-    (const char**)&data_brigade, 
-    (apr_size_t*)&len,
-    entryp
-    );
-
-  if (len > 0) {
-    DBG(r, "(in:exchange)POSTDATA:[%s]", data_brigade);
-
-    obb = apr_brigade_create(r->pool, c->bucket_alloc);
-
-    tmp_heap = apr_bucket_heap_create(data_brigade, 
-                                      len, 
-                                      NULL, 
-                                      f->c->bucket_alloc);
-    eos      = apr_bucket_eos_create(f->c->bucket_alloc);
-
-    APR_BRIGADE_INSERT_TAIL(obb, tmp_heap);
-    APR_BRIGADE_INSERT_TAIL(obb, eos);
-    APR_BRIGADE_CONCAT(bb, obb);
+  if (eos_flag) {
+    len = ctx->len;
+    data_brigade = chxj_input_convert(
+      r, 
+      (const char**)&ctx->buffer, 
+      (apr_size_t*)&len,
+      entryp
+      );
+  
+    if (len > 0) {
+      DBG(r, "(in:exchange)POSTDATA:[%s]", data_brigade);
+  
+      obb = apr_brigade_create(r->pool, c->bucket_alloc);
+  
+      tmp_heap = apr_bucket_heap_create(data_brigade, 
+                                        len, 
+                                        NULL, 
+                                        f->c->bucket_alloc);
+      eos = apr_bucket_eos_create(f->c->bucket_alloc);
+  
+      APR_BRIGADE_INSERT_TAIL(obb, tmp_heap);
+      APR_BRIGADE_INSERT_TAIL(obb, eos);
+      APR_BRIGADE_CONCAT(bb, obb);
+    }
+    else {
+      obb = apr_brigade_create(r->pool, c->bucket_alloc);
+      eos = apr_bucket_eos_create(f->c->bucket_alloc);
+      APR_BRIGADE_INSERT_TAIL(obb, eos);
+      APR_BRIGADE_CONCAT(bb, obb);
+    }
+    apr_pool_destroy(ctx->pool);
+    f->ctx = NULL;
   }
 
   DBG(r, "end of chxj_input_filter()");
@@ -1298,11 +1330,14 @@ chxj_create_per_dir_config(apr_pool_t *p, char *arg)
   conf->server_side_encoding = NULL;
   conf->cookie_db_dir    = NULL;
   conf->cookie_timeout   = 0;
+  conf->cookie_store_type = COOKIE_STORE_TYPE_NONE;
+  conf->cookie_lazy_mode  = 0;
 #if defined(USE_MYSQL_COOKIE)
   memset((void*)&conf->mysql, 0, sizeof(mysql_t));
   conf->mysql.port       = MYSQL_PORT;
   conf->mysql.host       = NULL;
-#elif defined(USE_MEMCACHE_COOKIE)
+#endif
+#if defined(USE_MEMCACHE_COOKIE)
   memset((void*)&conf->memcache, 0, sizeof(memcache_t));
   conf->memcache.host    = NULL;
   conf->memcache.port    = 0;
@@ -1510,8 +1545,8 @@ chxj_merge_per_dir_config(apr_pool_t *p, void *basev, void *addv)
   else {
     mrg->mysql.charset = NULL;
   }
-  
-#elif defined(USE_MEMCACHE_COOKIE)
+#endif
+#if defined(USE_MEMCACHE_COOKIE)
   if (add->memcache.host) {
     mrg->memcache.host = apr_pstrdup(p, add->memcache.host);
   }
@@ -1530,8 +1565,25 @@ chxj_merge_per_dir_config(apr_pool_t *p, void *basev, void *addv)
   else {
     mrg->memcache.port = 0;
   }
-
 #endif
+  if (add->cookie_store_type) {
+    mrg->cookie_store_type = add->cookie_store_type;
+  }
+  else if (base->cookie_store_type) {
+    mrg->cookie_store_type = base->cookie_store_type;
+  }
+  else {
+    mrg->cookie_store_type = COOKIE_STORE_TYPE_NONE;
+  }
+  if (add->cookie_lazy_mode) {
+    mrg->cookie_lazy_mode = add->cookie_lazy_mode;
+  }
+  else if (base->cookie_lazy_mode) {
+    mrg->cookie_lazy_mode = base->cookie_lazy_mode;
+  }
+  else {
+    mrg->cookie_lazy_mode = 0;
+  }
   return mrg;
 }
 
@@ -2139,7 +2191,8 @@ cmd_set_cookie_mysql_charset(
 
   return NULL;
 }
-#elif defined(USE_MEMCACHE_COOKIE)
+#endif
+#if defined(USE_MEMCACHE_COOKIE)
 static const char *
 cmd_set_cookie_memcache_port(
   cmd_parms   *UNUSED(cmd), 
@@ -2183,6 +2236,59 @@ cmd_set_cookie_memcache_host(
 }
 #endif
 
+static const char *
+cmd_set_cookie_lazy_mode(
+  cmd_parms   *UNUSED(cmd), 
+  void        *mconfig, 
+  const char  *arg)
+{
+  mod_chxj_config  *dconf;
+
+  if (strlen(arg) > 255) 
+    return "mod_chxj: ChxjCookieLazyMode is too long.";
+
+  dconf = (mod_chxj_config *)mconfig;
+
+  if (strcasecmp("TRUE",arg) == 0) {
+    dconf->cookie_lazy_mode = COOKIE_LAZY_ON;
+  }
+  else {
+    dconf->cookie_lazy_mode = COOKIE_LAZY_OFF;
+  }
+
+  return NULL;
+}
+
+static const char *
+cmd_set_cookie_store_type(
+  cmd_parms   *UNUSED(cmd), 
+  void        *mconfig, 
+  const char  *arg)
+{
+  mod_chxj_config  *dconf;
+
+  if (strlen(arg) > 255) 
+    return "mod_chxj: ChxjCookieStoreType is too long.";
+
+  dconf = (mod_chxj_config *)mconfig;
+
+  if (strcasecmp(CHXJ_COOKIE_STORE_TYPE_DBM, arg) == 0) {
+    dconf->cookie_store_type = COOKIE_STORE_TYPE_DBM;
+  }
+  else if (strcasecmp(CHXJ_COOKIE_STORE_TYPE_MYSQL, arg) == 0) {
+    dconf->cookie_store_type = COOKIE_STORE_TYPE_MYSQL;
+  }
+  else if (strcasecmp(CHXJ_COOKIE_STORE_TYPE_MEMCACHE, arg) == 0) {
+    dconf->cookie_store_type = COOKIE_STORE_TYPE_MEMCACHE;
+  }
+  else {
+    dconf->cookie_store_type = COOKIE_STORE_TYPE_NONE;
+  }
+
+  return NULL;
+}
+
+
 
 static const command_rec cmds[] = {
   AP_INIT_TAKE1(
@@ -2239,6 +2345,18 @@ static const command_rec cmds[] = {
     NULL,
     OR_ALL,
     "The compulsion time-out time of the cookie is specified. "),
+  AP_INIT_TAKE1(
+    "ChxjCookieStoreType",
+    cmd_set_cookie_store_type,
+    NULL,
+    OR_ALL,
+    "It specifies preserving of the cookie ahead. (DBM/MYSQL/MEMCACHE)"),
+  AP_INIT_TAKE1(
+    "ChxjCookieLazyMode",
+    cmd_set_cookie_lazy_mode,
+    NULL,
+    OR_ALL,
+    "OneTimeID is negligently done. (TRUE/FALSE)"),
 #if defined(USE_MYSQL_COOKIE)
   AP_INIT_TAKE1(
     "ChxjCookieMysqlHost",
@@ -2288,7 +2406,8 @@ static const command_rec cmds[] = {
     NULL,
     OR_ALL,
     "The MySQL charset used by saving Cookie"),
-#elif defined(USE_MEMCACHE_COOKIE)
+#endif
+#if defined(USE_MEMCACHE_COOKIE)
   AP_INIT_TAKE1(
     "ChxjCookieMemcacheHost",
     cmd_set_cookie_memcache_host,
